/*
 * anyRemote
 * a bluetooth remote for your PC.
 * Copyright (C) 2006
 * 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

/*
 * AT+CKPD commands handling
 *
 */

#include <stdlib.h>
 
#include "parse.h"
#include "common.h"

#define PAUSE_TIME 1

extern int serverMode;  // No AT command in serverMode
int useQuestionInCLCC = 0;

void sendSyncCommand(char *bytes, char *answer, int size, char *expect) 
{
    // initial timeout was 5
    bt_put_command (bytes, answer, size, 40, expect); 
}


// Send AT+CKPD command
static int sendCKPD(char *key)
{   
    char answer[1024];
    char ckpd[MAXCKPDLEN];
    
    if (key == NULL || key[0] == '\0') {
    	logger("DEBUG","Empty symbol in sendCKPD()");
        return EXIT_NOK;
    }
    sprintf(tmp,"sendCKPD >%s<", key);
    logger("DEBUG",tmp);
    
    sprintf(ckpd,"%s\"%s\"\r", DEF_AT_CKPD, key);
    
    int model = getModel();
    
    if (model == MODEL_SE || model == MODEL_SIEMENS) {	// SE-K700, Siemens-S55
    	// If we send AT+CKPD=X
	// SE phone will replay
	// OK
    	// +CKEV: X,1
    	// +CKEV: X,0
	// And wee need to filter out that CKEVs
	char expect[16];
	strcpy(expect, DEF_CKEV);
	strcat(expect, " ");
	strcat(expect, key);
	strcat(expect, ",0");  // Wait button up event
	
    	sendSyncCommand(ckpd, answer, sizeof(answer), expect);
    	sprintf(tmp,"Expect >%s<", expect);
    	logger("DEBUG",tmp);
    	sprintf(tmp,"Got all >%s<", answer);
    	logger("DEBUG",tmp);
	
	// Filter out button up
	char *ptr = strstr(answer, expect);
	if(ptr == NULL) {
	    logger("DEBUG","cant find button down event");
	    return EXIT_NOK;
	}
	int z;
	int l = strlen(expect);
	for (z=0; z<l; z++) {
	    *ptr = '\r';
	    ptr++;
	}
	// Filter out button down
	strcpy(expect, DEF_CKEV);
	strcat(expect, " ");
	strcat(expect, key);
	strcat(expect, ",1");
	ptr = strstr(answer, expect);
	if(ptr == NULL) {
	    logger("DEBUG","cant find button up event");
	    return EXIT_NOK;
	}
	l = strlen(expect);
	for (z=0; z<l; z++) {
	    *ptr = '\r';
	    ptr++;
	}
    	sprintf(tmp,"Filtered >%s<", answer);
    	logger("DEBUG",tmp);
	
	
    } else {
    	sendSyncCommand(ckpd, answer, sizeof(answer), NULL);
    }
    
    return handle_command(answer);
}


int sendSeq(char *seq) 
{
    if (getUseScreen() == 0) return EXIT_OK;

    if (seq == NULL) {
        logger("DEBUG","sendSeq() input is NULL");
        return EXIT_OK;
    }
    
    sprintf(tmp,"sendSeq() >%s<", seq);
    logger("DEBUG",tmp);
    
    if (serverMode > 0) {
    	logger("ERROR", "Can not use SendCKPD() in Server style");
	return EXIT_OK;
    } else {
        char copy[MAXARGLEN];
        strcpy(copy,seq);
	
        sprintf(tmp,"sendSeq() >%s<", seq);
        logger("DEBUG",tmp);
	char *bufPtr;
        char *ckpd = strtok_r(copy," ",&bufPtr);
    
        while (ckpd != NULL) {
           sprintf(tmp,"NEXT sendSeq()->sendCKPD()>%s<", ckpd);
           logger("DEBUG",tmp);
    	
	    if (strcmp(ckpd, PAUSE_STR)==0) {
	        // Just wait a little
    	        sprintf(tmp,"sendSeq() PAUSE");
    	        logger("DEBUG",tmp);
	    
	        sleep(PAUSE_TIME);
	    } else {
		
                int ret = sendCKPD(ckpd);
	        if (ret == EXIT_DISCON || ret == EXIT_ABORT) {
	    	    logger("WARNING", "Got disconnect or abort...");
	    	    return ret;
	        }
	    }
    	    ckpd = strtok_r(NULL," ",&bufPtr);
        }
    }
    return EXIT_OK;
}

int getClip(char* CallerID)
{
    char clcc[MTEXTLEN];
    char answer[1024];
    char *ptr;
    int   tryAnyway = 0;
    
    answer[0] = '\0';
    
    logger("DEBUG", "getClip()");
    
    
    // Some phones uses AT+CLCC?, some just AT+CLCC (Sagem, Siemens-S55, SE-K750)
    
    if (useQuestionInCLCC == 0) {
    	logger("DEBUG", "getClip() send AT+CLCC");
	
    	strcpy(clcc,DEF_AT_CLCC);
	strcat(clcc,"\r");
	
    	sendSyncCommand(clcc, answer, sizeof(answer), NULL);
    
    	if ((ptr = strstr(answer,DEF_CLCC)) != NULL) {
    	    strcpy(CallerID, parse_CLCC(ptr));
    	    return EXIT_EXACT;
    	} else if (strstr(answer,"OK") != NULL) {
    	    return EXIT_OK;
    	} else if (strstr(answer,"ERROR") != NULL) {
    	    logger("WARNING", "Got ERROR on AT+CLCC. Will try AT+CLCC?");
	    useQuestionInCLCC = 1; 	// Ask phone again; but will never ask AT+CLCC more
    	} else {
	    tryAnyway = 1;  	    	// Ask phone again; to be sure
	}
    }

    if (useQuestionInCLCC == 1 || tryAnyway == 1) {
    	
	tryAnyway = 0;
	
    	logger("DEBUG", "getClip() send AT+CLCC?");

    	strcpy(clcc,DEF_AT_CLCC);
    	strcat(clcc,"?");
	strcat(clcc,"\r");
	
    	sendSyncCommand(clcc, answer, sizeof(answer), NULL);
    
    	if ((ptr = strstr(answer,DEF_CLCC)) != NULL) {
    	    strcpy(CallerID, parse_CLCC(ptr));
    	    return EXIT_EXACT;
    	} else if (strstr(answer,"OK") != NULL) {
    	    return EXIT_OK;
    	} else if (strstr(answer,"ERROR") != NULL) {
    	    logger("ERROR", "Got ERROR on AT+CLCC?");
	    return EXIT_NOK;
    	}
    }
    
    sprintf(tmp,"sendClip() unappropriate answer: >%s<", answer);
    logger("DEBUG",tmp);
    
    return EXIT_NOK;
}
