/*
 * anyRemote
 * a bluetooth remote for your PC.
 * Copyright (C) 2006
 * 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */
 
#include <stdlib.h>
#include <time.h>
#include <sys/wait.h>
#include <sys/stat.h>

#include "atsend.h"
#include "cmds.h"
#include "common.h"
#include "parse.h"

extern void closePort(int final);

extern int  	    serverMode;
extern char 	    callerId[MAXLEN];
extern int  	    paramValue;
extern int  	    repeatNow;
extern type_key*    repeatCmdPtr;

char tmptext[MAXCMDLEN];
int  remoteOn       = 1;

timerCmd* timers  = NULL;

// This is not clean-up function. It just finished the connection.
static void finishAll()
{
    char cmd[32];
    char answer[1024];
	
    // finish events forwarding
    printf("Finish events forwarding\n");
    logger("INFO","Finish events forwarding");
    
    if (serverMode == 0) {
        strcpy(cmd,getAT_CMER(CMER_OFF));

        strcat(cmd,"\r");
        sendSyncCommand(cmd, answer, sizeof(answer), NULL);
    
        //handle_command(answer); // Does it needed ?
    
        // and resend No back
        //sendSyncCommand("AT+CKPD=\"e\"\r", answer, sizeof(answer), NULL);    // Does it needed ?
    }
    closePort(1);
    
    return;
}

static int execInFork(char *cmd)
{
    int ret;
    int cpid = fork();
    if( cpid < 0) {
    	logger("ERROR", "Can not fork!");
    	return -1;
    }

    if(cpid) {	    // father
    	waitpid(cpid,&ret,0);
    	if(ret!=-1 && WIFEXITED(ret)) {
	    ret = WEXITSTATUS(ret);
    	}
    } else {	    // child
    	logger("INFO", "Close port in child");
    	closePort(0);
	 
    	ret = system(cmd);
    	if(ret!=-1 && WIFEXITED(ret)) {
    	    ret = WEXITSTATUS(ret);
	}
    	exit(ret);
    }

    return ret;
}


static char *readResultsFile() 
{
    FILE *fp;

    fp=fopen(TOFILE,"r");
    if (fp == NULL) return NULL;
    
    int n = fread(tmptext, sizeof(char), MAXCMDLEN-1, fp);
    tmptext[n] = '\0';
    
    return tmptext;
}

static int execSimpleCmd(char *command, int toFile)
{
    char cmd [1024];
    cmd[0] = '\0';
    
    if (command == NULL) {
        logger("ERROR","execSimpleCmd(): null input");
        return EXIT_NOK;
    }
    //sprintf(tmp, "execSimpleCmd >%s<", command);
    //logger("INFO",tmp);
    
    if (*command == '\0') { // Empty command. Skip it.
    	return EXIT_OK;
    }
    
    if (remoteOn != 1) {
    	logger("DEBUG", "Remote if OFF. Skip event");
    	return EXIT_OK;
    }
    char *papameterized = NULL;
    
    int pc = strlen(PARAM_CALLID);
    int pt = strlen(PARAM_TIME);
    int pm = strlen(PARAM_MODE);
    int pp = strlen(PARAM_PARAM);
     
    while(command != NULL) {
    	sprintf(tmp, "cur cmd >%s<", command);
    	logger("DEBUG",tmp);
	
        papameterized = strstr(command, PARAM_START);   // $(
 
        sprintf(tmp, "papameterized >%s<", papameterized);
        logger("DEBUG",tmp);
        
	if (papameterized == NULL) { // Just copy
	    strcat(cmd, command);
            break;
        } else if(strncmp(papameterized+2,PARAM_CALLID,pc) == 0) {

            *papameterized = '\0';
            
	    // First part
    	    strcat(cmd, command);
	    
	    // Add params
	    strcat(cmd, callerId);
            
            command = papameterized + pc + 2;
	    
	} else if (strncmp(papameterized + 2,PARAM_TIME,pt) == 0) { 
	    
            *papameterized = '\0';

	    // First part
            strcat(cmd, command);
	    
	    // Add time
	    time_t logtime;
	    time(&logtime);
	
	    struct tm curtime;
	    localtime_r(&logtime, &curtime);
	
	    char buf[64];
	    sprintf(buf,"%d:%02d", curtime.tm_hour, curtime.tm_min);
	
	    strcat(cmd, buf);
 
            command = papameterized + pt + 2;
	    
        } else if (strncmp(papameterized + 2,PARAM_MODE,pm) == 0) {
	   
            *papameterized = '\0';

	    // First part
    	    strcat(cmd, command);
	    
	    // Add params
	    strcat(cmd, getMode());
		    
            command = papameterized + pm + 2;
	    
        } else if (strncmp(papameterized + 2,PARAM_PARAM,pp) == 0) {
            // This is dirty trick to make Bemused commands like VOLM<volume> works
            
            *papameterized = '\0';

            if (paramValue == -1) {
                logger("ERROR", "Parameter value was not set!");
                return EXIT_OK;
            }
	    
	    // First part
    	    strcat(cmd, command);

 	    // Add params
	    char buf[64];
	    sprintf(buf, "%d", paramValue);
	    strcat(cmd, buf);
            
            paramValue = -1; // Reset value after use
            
            command = papameterized + pp + 2;
	    
        } else {    // Just copy
	    strcat(cmd, command);
            break;
	}
        sprintf(tmp, "partial cmd >%s<", cmd);
        logger("DEBUG",tmp);
    }
    sprintf(tmp, "final cmd >%s<", cmd);
    logger("DEBUG",tmp);

    if (toFile) {       // We needs to wait results, so do not run in background
        strcat(cmd, " > ");
        strcat(cmd, TOFILE);
    } else {
        if (! index(cmd, '&')) {
            strcat(cmd, "&");
        }
    }
    return (/*system(cmd)*/execInFork(cmd) == -1 ? EXIT_NOK : EXIT_OK);
}

static int processSendCmd(char *origCmd)
{
    int ret = EXIT_OK;
    
    char all[MAXMAXLEN];
    strncpy(all, origCmd, MAXMAXLEN-1);
    
    char *tag  = strtok(all, ",");
    if (tag == NULL) {
        logger("DEBUG", "processSendCmd(): incorrect input (1)");                
        return EXIT_NOK;
    }
    if (strlen(origCmd) <= strlen(tag) + 1) {   // Set(tag,) or Set(tag) without any value
        logger("DEBUG", "processSendCmd(): incorrect input (2)");                
        return EXIT_NOK;
    }
    char *ptr   = tag + strlen(tag) - 1;
    // strip spaces from tail
    while( *ptr == '\0' || *ptr == '\n' || *ptr == '\t' || *ptr == ' ') {
    	*ptr = '\0';
        ptr--;
    }
    ptr = tag + strlen(tag) + 1;
    // strip spaces from head of value
    while( *ptr != '\0' && (*ptr == '\t' || *ptr == ' ')) {
        ptr++;
    }
    
    sprintf(tmp, "processSendCmd() is >%s< >%s<",tag,ptr);
    logger("DEBUG", tmp);
    
    if (strcmp(tag, SEND_STRING) == 0) {
        ret = writeSocketConn(ptr);
        if (ret != EXIT_OK) {
	    logger("DEBUG","Fails in writeSocketConn()");
        }
    } else if (strcmp(tag, SEND_BYTES) == 0) {
        ret = writeBytesSocketConn(ptr);
        if (ret != EXIT_OK) {
	    logger("DEBUG","Fails in writeBytesSocketConn()");
        }    
    } else {
        logger("DEBUG", "processSendCmd(): incorrect input (3)");                
        return EXIT_NOK;
    }
    return ret;
}

static int execCmdAndGetResults(int type, char *commandIn, timerCmd* timer) 
{
    logger("DEBUG","execCmdAndGetResults");

    if (serverMode != SERVER_NO && remoteOn == 1 && commandIn != NULL) {
        
        // Parse type of set
        char commandOrig[MAXMAXLEN];
        strncpy(commandOrig, commandIn, MAXMAXLEN-1);
        
        // commandIn = tag,command
        char* tag = strtok(commandOrig,",");
        char* ttl = NULL;
        int commas = 1;
        int tlen   = 0;
        int glen   = 0;

        if (tag == NULL) {
            logger("DEBUG", "execCmdAndGetResults(): incorrect input (1)");                
            return EXIT_NOK;
        }
        
        glen = strlen(tag);
        
        if (strstr(tag,SET_TEXT) != NULL) { // text,title,cmd
            ttl = strtok(NULL,",");
            if (ttl != NULL) {
                tlen = strlen(ttl);
            }
            commas = 2;
        }
        sprintf(tmp, "execCmdAndGetResults() is >%s< >%s<",tag,commandIn);
        logger("DEBUG", tmp);
        
        if (strlen(commandIn) <= glen + tlen + commas) {   // Set(tag,) or Set(tag) without any value
            logger("DEBUG", "execCmdAndGetResults(): incorrect input (2)");                
            return EXIT_NOK;
        }
        if (type != ID_SET && type != ID_SEND ) {
            logger("DEBUG", "execCmdAndGetResults(): incorrect input (3)");                
            return EXIT_NOK;
        }
        char *command2exec = tag + glen + tlen + commas;
        
        while(*tag == ' ' || *tag == '\t') {
            tag++;
        }
        
        int ret = execSimpleCmd(command2exec,1);
        if (ret != EXIT_OK) {
 	    sprintf(tmp, "Return code from execSimpleCmd() is %d",ret);
 	    logger("DEBUG", tmp);
	    return ret;
        }
        char *res = readResultsFile();
	char *res2;
        int  braces = 1;
        
        if (type == ID_SET) {
            res2  = calloc(glen + tlen + strlen(res) + strlen(CMD_SET) + commas + 4, 1);
	    strcpy(res2,CMD_SET);
        } else {                        // ID_SEND - Just send string
            res2  = calloc(strlen(res)+ 1, 1);
            braces = 0;
        }
	
        if (braces) {
            strcat(res2,"(");
            strcat(res2,tag);
            strcat(res2,",");
            if (ttl != NULL) {
                strcat(res2,ttl);
                strcat(res2,",");
            }
        }
	strcat(res2,res);
        if (braces) {
	    strcat(res2,");");
        }
	res[0] = '\0';
	
        int needToWrite = 1;
        if (timer != NULL && timer->lastValue != NULL && strcmp(timer->lastValue,res2) == 0) {
            logger("DEBUG","No needs to send the same data.");
            needToWrite = 0;            // no needs to send exactly the same string
        }
        if (needToWrite) {
            if (type == ID_SEND && strcmp(tag,SEND_BYTES) == 0) {
	        ret = writeBytesSocketConn(res2);
            } else {
	        ret = writeSocketConn(res2);
            }
            if (ret != EXIT_OK) {
	        logger("DEBUG","Fails in write(Bytes)SocketConn()");
            }
            // Store last results
            if (timer != NULL) {
                if (timer->lastValue == NULL) {
                    timer->lastValue = res2;
                } else {
                    free(timer->lastValue);
                    timer->lastValue = res2;
                }
            } else {
                free(res2);res2 = NULL;
            }
        }
        if (timer == NULL) {
            free(res2);res2 = NULL;
        }
    }
       
    return EXIT_OK;
}

static int processOneCommand(int cmdType, char *commandIn)
{
    int ret;
    char commandCopy[MAXMAXLEN];
    char *commandOrig = NULL;

    if (commandIn == NULL) {
        sprintf(tmp, "execCmd >no params< >%d<", remoteOn);
    } else {
        sprintf(tmp, "execCmd >%s< >%d<", commandIn, remoteOn);
        strncpy(commandCopy, commandIn, MAXMAXLEN-1);
        commandOrig = commandCopy;
    }
    logger("INFO",tmp);
    
    
    switch (cmdType) {
    
        case ID_EXIT:               
    	    if (remoteOn == 1) {	    // User asks to exit	
    	        logger("INFO", "Exit key detected");
    	        finishAll();			    

    	        return EXIT_ABORT;
            }
            break;
            
        case ID_TMPDISCONN:
            if (remoteOn == 1) {            // User asks to temporary disconnect
    	        logger("INFO", "Temporary disconnect key detected");
	        finishAll();
                
    	        return EXIT_DISCON; // But what with the rest of the command line ? (now just skip)
            }
            break;
            
        case ID_REMOTEON:
            logger("INFO", "Switch remote on");
    	    remoteOn = 1;
            break;
            
        case ID_REMOTEOFF:
            logger("INFO", "Switch remote off");
    	    remoteOn = 0;
            break;
            
        case ID_EXEC:
            logger("INFO", "Command: Exec");
            if (remoteOn == 1) {
                execSimpleCmd(commandOrig, 0);
            }
            break;
            
        case ID_SETMODE:
            logger("INFO", "Command: SetMode");
            if (remoteOn == 1) {
	    
	    	int i = 0;    
	    	type_key* k = findItem(EVT_EXITMODE, &i);
    	    	if (k && i == 1) {
    	            logger("DEBUG", "Exec cmd on exit from mode");
    	            ret = handleCmdByKey(k);
    	            if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
    	    	    	return ret;
    	            }
    	    	} else {
    	            logger("DEBUG", "Leave mode");
    	    	}
                setMode(commandOrig);
		
		i = 0; 
	    	k = findItem(EVT_ENTRMODE, &i);
    	    	if (k && i == 1) {
    	            logger("DEBUG", "Exec cmd on enter to mode");
    	            ret = handleCmdByKey(k);
    	            if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
    	    	    	return ret;
    	            }
    	    	} else {
    	            logger("DEBUG", "Entered to mode");
    	    	}
            } else {
    	            logger("DEBUG", "remote Off ???");
    	    }
            break;

        case ID_SENDCKPD:
            logger("INFO", "Command: SendCKPD");
            if (serverMode == SERVER_NO && remoteOn == 1) {
 
                ret = sendSeq(commandOrig);
                if (ret != EXIT_OK) {
	            logger("DEBUG","Fails in sendSeq()");
                }
                if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
    	            return ret;
                }
            }    
            break;

        case ID_SEND:
            logger("INFO", "Command: Send");
            if (serverMode != SERVER_NO && remoteOn == 1 && commandOrig != NULL) {
                ret = processSendCmd(commandOrig);
                if (ret != EXIT_OK) {
	            logger("DEBUG","Fails in processSendCmd()");
                }
            }    
            break;
            
            
        case ID_SET:          
            logger("INFO", "Command: Set");
            if (serverMode != SERVER_NO && remoteOn == 1 && commandOrig != NULL) {
                strcpy(tmptext, CMD_SET);
                strcat(tmptext, "(");
                
                // Change \n in input to real '\n'
                char *p1 = commandOrig;
                char *p2 = commandOrig;
                while (p2 != NULL) {
                    p2 = strstr(p1, "\\n");
                    
                    if (p2 == NULL) {
                        strncat(tmptext, p1, p2-p1);
                    } else {
                        strncat(tmptext, p1, p2-p1);
                        strcat(tmptext, "\n");
                    }
                    p1 = p2 + 2;
                }
                //strcat(tmptext, commandOrig);
                strcat(tmptext, ");");
               
                ret = writeSocketConn(tmptext);
                if (ret != EXIT_OK) {
	            logger("DEBUG","Fails in writeSocketConn() (Set)");
                }
            }    
            break;
            
        case ID_EXECSET:   
            logger("INFO", "Command: ExecAndSet");
            
            execCmdAndGetResults(ID_SET, commandOrig, NULL);
            break;
            
        case ID_EXECSEND:  
            logger("INFO", "Command: ExecAndSend");
            
            execCmdAndGetResults(ID_SEND, commandOrig, NULL);
            break;
            
        case ID_TIMERCMD:   
            logger("INFO", "Command: Timer");
            if (serverMode != SERVER_NO && remoteOn == 1 && commandOrig != NULL) {
                return processTimerCmd(commandOrig);
            }     
            break;

        case ID_CANCELTIMER:   
            logger("INFO", "Command: CancelTimer");
            if (serverMode != SERVER_NO && remoteOn == 1 && commandOrig != NULL) {
                cancelTimer(commandOrig);
            }  
            break;
            
        default:  
            logger("INFO", "Unknown command");                 
            break;
   }
   
   return EXIT_OK;
}

int processCommands(cmdItem* cmdList) 
{
    while (cmdList) {
    	if (cmdList->descr == NULL) {
		sprintf(tmp, "Process next command NULL parameters");
	} else {
        	sprintf(tmp, "Process next command %s",cmdList->descr);
        }
	logger("DEBUG", tmp);
            
        int ret = processOneCommand(cmdList->type, cmdList->descr);
        if (ret != EXIT_OK) {
            sprintf(tmp, "Return code from processOneCommand() is %d",ret);
            logger("DEBUG", tmp);
 	     
            if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
  	        logger("DEBUG", "Abort or disconnect after processOneCommand()");
		
	        return ret;
            }
        }
        cmdList = cmdList->next;
    }
    return EXIT_OK;
}

//
// Top-level command processing function
//
int handleCmdByKey(type_key* k)
{
    if (k == NULL) {
    	logger("DEBUG", "handleCmdByKey() NULL input");
	return EXIT_OK;
    }

    sprintf(tmp, "handleCmdByKey() >%s<",k->key);
    logger("DEBUG", tmp);
    
    cmdItem* cmdList = getCommand(k);
    if (cmdList == NULL) {
    	logger("DEBUG", "No appropriate command was found by getCommand()");
    } else {
     	sprintf(tmp, "ready to execute command");
     	logger("DEBUG", tmp);
        
        // In AT mode: User had pressed some key ... 
        // Send ToMainMenu sequence of CKPD's to show main menu again
        if (serverMode == SERVER_NO) {
            if (sendSeq(getToMainMenu()) != EXIT_OK) {
    	        logger("DEBUG","Fails in sendSeq()");
            }
        }
        
        int ret = processCommands(cmdList);
    	if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
 	    repeatNow    = 0;
 	    repeatCmdPtr = NULL;
	    return ret;
        }
    }	
	
    return EXIT_OK;	
}

//
// Functions related to timer commands
//
// Timer(id,timeout,maxtimes,tag,command);
//      where tag could be: icons, title, status
//

timerCmd* findTimer(char *id, timerCmd **prev) 
{
    timerCmd* ptr = timers;
    
    while (ptr != NULL && strncmp(ptr->id,id,MTEXTLEN) != 0) {
        *prev = ptr;
        //sprintf(tmp, "findTimer(): set prev to >%s<", (*prev)->id);
        //logger("DEBUG",tmp);
        ptr  = ptr->next;
    }

    return ptr;
}

static int createTimer(char* id, int type, float tmout, int times, char *cmd)
{
    if (id == NULL || cmd == NULL) {
        logger("DEBUG", "createTimer(): null input");                
        return EXIT_NOK;
    }
    timerCmd* ptr = NULL;
    if (findTimer(id, (timerCmd**) &ptr) != NULL) {
        logger("DEBUG", "createTimer(): timer already exists");                
        return EXIT_NOK;
    } 
    timerCmd *tm = calloc(1,sizeof(timerCmd));
    
    strncpy(tm->id, id, MTEXTLEN);
    tm->type      = type;
    tm->ticks     = 0;
    tm->times     = 1;
    tm->tmout     = tmout;
    tm->maxTimes  = times;
    tm->lastValue = NULL;
    
    tm->cmd = (char *) calloc(1, strlen(cmd) + 1);
    strcpy(tm->cmd, cmd);
    
    // Insert in head
    if (timers == NULL) { // Insert first
        tm->next = NULL;
    } else {
        tm->next = timers;
    }
    timers = tm;
    
    return EXIT_OK;
}

int cancelTimer(char *id) 
{
    sprintf(tmp, "cancelTimer() >%s<", id);
    logger("DEBUG", tmp);
    
    timerCmd *prev = NULL;
    timerCmd *tm = findTimer(id, (timerCmd **) &prev);
    if (tm == NULL) {
        logger("DEBUG", "Cant find timer to cancel!");
        return EXIT_NOK;
    }    
    if (tm == timers) { //remove in head
        timers = tm->next;
    } else {
        if (prev != NULL) {
            prev->next = tm->next;
        } else {
            logger("ERROR", "Previous item absent for non-first element");
            return EXIT_NOK;
        }
    }        

    free(tm->cmd); tm->cmd = NULL;
    free(tm);
    
    return EXIT_OK;
}


int processTimerCmd(char *origCmd)
{
    char all[MAXMAXLEN];
    strcpy(all, origCmd);
    
    char *id  = strtok(all, ",");
    if (id == NULL) {
        logger("DEBUG", "processTimerCmd(): incorrect input - timer data absent");                
        return EXIT_NOK;
    }

    char *ptr = id + strlen(id) - 1;
    // strip spaces from tail
    while( *ptr == '\0' || *ptr == '\n' || *ptr == '\t' || *ptr == ' ') {
    	*ptr = '\0';
        ptr--;
    }
    
    char* t1  = strtok(NULL, ",");
    char* t2  = strtok(NULL, ",");
    char* cmd = strtok(NULL, "\0");  
    if (t1 == NULL || t2 == NULL || cmd == NULL) {
        logger("DEBUG", "processTimerCmd(): incorrect input - command is not full");                
        return EXIT_NOK;
    }
    while (*cmd == ' ' || *cmd == '\t') {   // skip spaces
        cmd++;
    }
    sprintf(tmp, "processTimerCmd() cmd = >%s<", cmd);
    logger("DEBUG", tmp);
    
    // Do not allow timer commands for Send(string/bytes...)
    if (strncmp(cmd,SET_ICONS,  strlen(SET_ICONS))  != 0 &&
        strncmp(cmd,SET_TITLE,  strlen(SET_TITLE))  != 0 &&
        strncmp(cmd,SET_STATUS, strlen(SET_STATUS)) != 0 &&
        strncmp(cmd,SET_STATUS, strlen(SET_TEXT))   != 0) {
        logger("DEBUG", "processTimerCmd(): incorrect tag in command");                
        return EXIT_NOK;
    }
    
    float tmout = (float) atof(t1);
    int   times = atoi(t2);

    if (createTimer(id, ID_SET, tmout, times, cmd) == EXIT_NOK) {
        logger("DEBUG", "processTimerCmd(): timer already exists or input was incorrect");                
        return EXIT_NOK;
    }
    
    timerCmd* tt = NULL;
    execCmdAndGetResults(ID_SET, cmd, findTimer(id, &tt));

    return EXIT_OK;
}

void freeTimers()
{
    logger("DEBUG", "freeTimers()");
    timerCmd *tm;
    
    while (timers != NULL) {
        tm     = timers;
        timers = timers->next;
        
        free(tm->cmd); tm->cmd = NULL;
        free(tm);
    }
    return;
}

int verifyTimers(int ticksInSec) 
{
    timerCmd *tm   = timers;
    //timerCmd *prev = NULL;
    
    int numTimers = 0;
    while (tm != NULL) {
        
        if (tm->ticks >= (tm->tmout * ticksInSec)) {
            sprintf(tmp, "verifyTimers(): it is time to execute >%s<", tm->cmd);
            logger("DEBUG",tmp);

            execCmdAndGetResults(tm->type, tm->cmd, tm);
            
            tm->ticks = 0;
            
            if (tm->maxTimes > 0) {
                tm->times++;
                if (tm->times >= tm->maxTimes) {    // Cancel this timer 
                    sprintf(tmp, "verifyTimers(): timer %s self-canceled", tm->id);
                    logger("DEBUG", tmp);
                    
                    // Cancel this timer
                    timerCmd *tmc = tm;
                    
                    // Relink timers list
                    //if (tm == timers) {  // First one
                    //    timers = timers->next;
                    //} else {
                    //    prev->next = tm->next;
                    //}
                    
                    tm = tm->next;

                    cancelTimer(tmc->id);
                    
                    continue;
                }
            }
        } else {
            tm->ticks++;
        }
        //prev = tm;
        tm   = tm->next;
        numTimers++;
    }
    return numTimers;
}
